#version 150

//different operations that can be applied to the heightmap
#define MODE_CLEAR 0
#define MODE_SMOOTH 1
#define MODE_PERLIN 2
#define MODE_FRACTAL 3
#define MODE_NORMALS 4
#define MODE_NOISE 5
#define MODE_FAULT 6
#define MODE_COLOUR 7
#define MODE_AMBIENT 8

uniform int mode; //one of the above enums

//shader output: new heights, colours or normals
out vec4 fragColour;

//heightmap dimensions
uniform ivec2 size;

//texel size
vec2 p = vec2(1.0) / size;

//input maps
uniform sampler2D height;
uniform sampler2D normal;
uniform sampler2D noise;

//used for many operations
uniform float strength;

//for blur
uniform vec2 direction;

//for perlin noise
uniform int levels;

//for the fractal method
uniform int step;
uniform int square;

//for the fault method
uniform vec2 faultPos;
uniform vec2 faultDir;

//for colouring
uniform vec3 splatColour;
uniform vec2 splatHeight;
uniform vec2 splatAngle;
uniform float splatCoverage;
uniform vec2 splatStart;
uniform vec2 splatEnd;

float noise2D(vec2 pos)
{
	pos = mod(pos, vec2(1.0));
	return texture(noise, pos).r;
}

float perlin2D(vec2 pos, int levels)
{
	float total = 0.0;
	float p = 0.5;
	int n = levels - 1;
	for (int i = 0; i < n; ++i)
	{
		float frequency = pow(2.0, float(i));
		float amplitude = pow(p, i);
		total += noise2D(pos * frequency / 100.0) * amplitude;
	}
	return total;
}

float blur(vec2 coord)
{
	const int r = 10;
	vec2 c = coord;
	float total = 0.0;
	float average = 0.0;
	float weight = 1.0;
	for (int i = -r; i <= r; ++i)
	{
		c += direction * p;
		float h = texture(height, c).r;
		average += h * weight;
		total += weight;
		weight *= strength;
	}
	average /= total;
	return average;
}

bool accumHeight(vec2 coord, inout float average)
{
	if (coord.x < 0.0 || coord.x > step/float(size.x) ||
		coord.y < 0.0 || coord.y > step/float(size.y))
		return false;
	average += texture(height, coord).r;
	return true;
}

//this seems to work, but I would not be surprised if it has bugs.
float diamondSquare(ivec2 icoord, vec2 coord)
{
	float random = texture(noise, coord * 64.234).r - 0.5;
	
	//set first iteration of noise
	if (step == 2 && square == 1)
	{
		return random * strength;
	}
	
	vec2 oldcoord = vec2(icoord/ivec2(2))/vec2(size) + p * 0.5;
	if (square == 1)
	{
		float last = texture(height, coord).r;
		if ((icoord.x + icoord.y) % 2 == 0) return last;
	}
	else
	{
		float last = texture(height, oldcoord).r;
		if (icoord.x % 2 == 0) return last;
		if (icoord.y % 2 == 0) return last;
	}
	
	float average = 0.0;
	float total = 0.0;
	if (square == 1)
	{
		if (accumHeight(coord + vec2(p.x, 0), average)) total += 1.0;
		if (accumHeight(coord + vec2(-p.x, 0), average)) total += 1.0;
		if (accumHeight(coord + vec2(0, p.y), average)) total += 1.0;
		if (accumHeight(coord + vec2(0, -p.y), average)) total += 1.0;
	}
	else //diamond
	{
		if (accumHeight(oldcoord + vec2(p.x, p.y), average)) total += 1.0;
		if (accumHeight(oldcoord + vec2(p.x, 0.0), average)) total += 1.0;
		if (accumHeight(oldcoord + vec2(0.0, 0.0), average)) total += 1.0;
		if (accumHeight(oldcoord + vec2(0.0, p.y), average)) total += 1.0;
	}
	average /= total;
	return average + random * strength;
}


//don't read too much into this. it's quite messy and has no real structure
//just me playing with different functions
vec4 makeColour(vec2 coord)
{
	vec4 colour;
	vec2 noiseCoord = coord;
	noiseCoord += mod(splatHeight * 10.0, 1.0);
	noiseCoord += mod(splatAngle * 10.0, 1.0);
	float random = perlin2D(noiseCoord * 50.0, 6) - 0.5;
	float randomLarge = perlin2D(vec2(0.4) + noiseCoord * 4.0, 6) - 0.5;
	float c = splatCoverage;
	
	float height = texture(height, coord).r;
	vec3 normal = normalize(texture(normal, coord).rgb * 2.0 - 1.0);
	float d = (height - splatHeight.x) / (splatHeight.y - splatHeight.x);
	
	float nearHeight = pow(1.0 - abs(d * 2.0 - 1.0), 0.4);
	c -= (1.0 - nearHeight) * max(0.0, sign(randomLarge + splatCoverage)) * 10.0 * (1.0 - c);
	c *= nearHeight;
	
	float cosa = dot(normal, vec3(0, 0, 1));
	float e = (cosa - splatAngle.x) / (splatAngle.y - splatAngle.x);
		
	c *= pow(1.0 - abs(e * 2.0 - 1.0), 0.1);
		
	float rscale = pow(1.0 - abs(c * 2.0 - 1.0), 0.1);
	colour.a = c + random * rscale;

	colour.rgb = splatColour;
	return colour;
}

//self explanatory - branches to different functions depending on mode
void main()
{
	ivec2 icoord = ivec2(gl_FragCoord.xy);
	vec2 coord = vec2(gl_FragCoord.xy) / vec2(size);
	switch (mode)
	{
	case MODE_CLEAR:
		fragColour = vec4(0,0,0,1);
		break;
	case MODE_SMOOTH:
		fragColour = vec4(blur(coord),0,0,1);
		break;
	case MODE_PERLIN:
		fragColour = vec4((perlin2D(coord * 16.0, levels)-0.5)/2.0,0,0,1);
		break;
	case MODE_FRACTAL:
		fragColour = vec4(diamondSquare(icoord, coord),0,0,1);
		break;
	case MODE_NORMALS:
		const int r = 1;
		float dx = 1.0;
		float dy = 0.0;
		float o = texture(height, coord).r;
		for (int i = -r; i <= r; ++i)
		{
			for (int j = -r; j <= r; ++j)
			{
				vec2 s = coord + p * vec2(i, j);
				float h = texture(height, s).r - o;
				if (i != 0 && s.x >= 0.0 && s.x <= 1.0)
					dx += h / (i * p.x);
				if (j != 0 && s.y >= 0.0 && s.y <= 1.0)
					dy += h / (j * p.y);
			}
		}
		float n = pow(r * 2.0, 2.0);
		dx /= n;
		dy /= n;
		vec3 norm = vec3(-dx, -dy, 10.0);
		fragColour = vec4(normalize(norm) * 0.5 + 0.5,1);
		break;
	case MODE_NOISE:
		float random = texture(noise, coord).r;
		fragColour = vec4(random * strength, 0, 0, 1);
		break;
	case MODE_FAULT:
		vec2 coordDir = coord - faultPos;
		if (cross(vec3(coordDir, 0.0), vec3(faultDir, 0.0)).z > 0.0)
			fragColour = vec4(strength*0.5, 0, 0, 1);
		else
			fragColour = vec4(-strength*0.5, 0, 0, 1);
		break;
	case MODE_COLOUR:
		vec2 oc = coord;
		coord *= splatEnd - splatStart;
		coord += splatStart;
		fragColour = makeColour(coord);
		break;
	}
}
